"""Kukur makes time series data and metadata available to the Apache Arrow ecosystem."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import typing

from datetime import datetime
from typing import Generator, Protocol, Union

import pyarrow as pa

from .base import (
    DataType,
    Dictionary,
    InterpolationType,
    Metadata,
    ProcessType,
    SeriesSelector,
)
from .client import Client


@typing.runtime_checkable
class Source(Protocol):
    """Source is the interface that Kukur data sources need to implement."""

    def search(
        self, selector: SeriesSelector
    ) -> Generator[Union[SeriesSelector, Metadata], None, None]:
        """Return all time series or even the metadata of them in this source matching the selector."""
        ...

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Return metadata for the given time series."""
        ...

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data for the given time series in the given time period."""
        ...


__all__ = [
    "Client",
    "DataType",
    "Dictionary",
    "InterpolationType",
    "Metadata",
    "ProcessType",
    "SeriesSelector",
]



"""The main objects in Kukur."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import typing

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Union


@dataclass
class Dictionary:
    """A Dictionary maps numbers to labels.

    Time series can contain integer values that have a meaning. For example
    the number 0 could be 'OFF' and 1 'ON'.

    The ordering of the labels can also be significant while presenting the
    series to a user.

    In Python 3.8+, iteration over a dict keeps the insert ordering."""

    mapping: Dict[int, str]


@dataclass
class SeriesSelector:
    """SeriesSelector identifies a group of time series matching the given pattern."""

    source: str
    name: Optional[str] = None


class InterpolationType(Enum):
    """InterpolationType describes how the value of a series evolves between data points."""

    LINEAR = "LINEAR"
    STEPPED = "STEPPED"


class DataType(Enum):
    """DataType represents the data type of the values in a time series.

    FLOAT32 is a 32 bit floating point number
    FLOAT64 is a 64 bit floating point number
    STRING is a variable length utf-8 character array
    DICTIONARY is an enumeration of ordered discrete values with a corresponding mapping
    CATEGORICAL is an enumeration of ordered discrete values
    """

    FLOAT32 = "FLOAT32"
    FLOAT64 = "FLOAT64"
    STRING = "STRING"
    DICTIONARY = "DICTIONARY"
    CATEGORICAL = "CATEGORICAL"


class ProcessType(Enum):
    """ProcessType represents the process type of an analysis"""

    CONTINUOUS = "CONTINUOUS"
    REGIME = "REGIME"
    BATCH = "BATCH"
    COUNTER = "COUNTER"


@dataclass
class Metadata:  # pylint: disable=too-many-instance-attributes
    """Metadata contains metadata fields known by Kukur."""

    series: SeriesSelector
    description: str = ""
    unit: str = ""
    limit_low: Optional[float] = None
    limit_high: Optional[float] = None
    accuracy: Optional[float] = None
    interpolation_type: Optional[InterpolationType] = None
    data_type: Optional[DataType] = None
    dictionary_name: Optional[str] = None
    dictionary: Optional[Dictionary] = None
    process_type: Optional[ProcessType] = None

    __json_mapping = {
        "limit_low": "limitLow",
        "limit_high": "limitHigh",
        "interpolation_type": "interpolationType",
        "data_type": "dataType",
        "dictionary_name": "dictionaryName",
        "dictionary": "dictionary",
        "process_type": "processType",
    }

    __human_mapping = {
        "lower limit": "limit_low",
        "upper limit": "limit_high",
        "interpolation type": "interpolation_type",
        "data type": "data_type",
        "dictionary name": "dictionary_name",
        "process type": "process_type",
    }

    __hidden = [
        "dictionary name",
        "dictionary",
    ]

    def __iter__(self):
        reverse_mapping = {v: k for k, v in self.__human_mapping.items()}
        for k, v in vars(self).items():
            if k == "series":
                continue
            yield (reverse_mapping.get(k, k), self._serialize_value(v))

    def camelcase(self):
        """Converts this Metadata instance to a dict that follows the camelCase naming convention.

        This dict is suitable for JSON."""
        return {
            self.__json_mapping.get(k, k): self._to_json(v)
            for k, v in vars(self).items()
        }

    def set_field(self, name: str, value: Union[Optional[float], str, Dictionary]):
        """Set a metadata field from the human friendly or camelcase field key representation."""
        mapped_name = self.__human_mapping.get(name, name)
        if mapped_name not in vars(self):
            camelcase_mapping = {v: k for k, v in self.__json_mapping.items()}
            mapped_name = camelcase_mapping.get(name, name)
            if mapped_name not in vars(self):
                raise AttributeError()

        hints = typing.get_type_hints(self.__class__)
        options = typing.get_args(hints[mapped_name])
        if isinstance(value, Dictionary):

            def convert_fn(value):
                return value

        elif len(options) == 0:
            convert_fn = hints[mapped_name]
        else:
            none_class = None.__class__
            convert_fn = [clazz for clazz in options if clazz != none_class][0]

        vars(self)[mapped_name] = convert_fn(value)

    def get_field(self, name: str) -> Union[Optional[float], str]:
        """Get the value of the metadata field with the given human friendly name."""
        mapped_name = self.__human_mapping.get(name, name)
        return vars(self)[mapped_name]

    def is_shown(self, name: str) -> bool:
        """Determine if the field with the given human friendly name should be displayed.

        The 'dictionary' will only be shown when it contains actual values for example.
        """
        return name not in self.__hidden or (
            self.get_field(name) is not None and self.get_field(name) != ""
        )

    def _to_json(self, v: Any):
        if isinstance(v, Dictionary):
            return list(v.mapping.items())
        return self._serialize_value(v)

    def _serialize_value(self, v: Any):  # pylint: disable=no-self-use
        if isinstance(v, Enum):
            return v.value
        return v



"""Connect to another Kukur instance using Arrow Flight."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Any, Dict, Generator, Tuple, Union

import pyarrow as pa

from kukur import Metadata, SeriesSelector
from kukur.client import Client
from kukur.exceptions import InvalidDataError


class TimeseerNotInstalledError(Exception):
    """Raised when the timeseer is not available."""

    def __init__(self):
        Exception.__init__(self, "the timeseer is not available. Install timeseer.")


def from_config(config: Dict[str, Any]):
    """Create a new Kukur source"""
    source_name = config["source"]
    host = config.get("host", "localhost")
    port = config.get("port", 8081)
    api_key = (config.get("api_key_name", ""), config.get("api_key", ""))
    return KukurSource(source_name, host, port, api_key)


class KukurSource:
    """KukurSource connects to another Kukur instance using Arrow Flight."""

    __client: Client
    __source_name: str

    # pylint: disable=too-many-arguments
    def __init__(
        self, source_name: str, host: str, port: int, api_key: Tuple[str, str]
    ):
        self.__client = Client(api_key, host, port)
        self.__source_name = source_name

    def search(
        self, selector: SeriesSelector
    ) -> Generator[Union[Metadata, SeriesSelector], None, None]:
        """Search time series using the Flight service."""
        query = SeriesSelector(self.__source_name, selector.name)
        for result in self.__client.search(query):
            if isinstance(result, SeriesSelector):
                yield SeriesSelector(selector.source, result.name)
            else:
                result.series = SeriesSelector(selector.source, result.series.name)
                yield result

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Get metadata from the Flight service."""
        if selector.name is None:
            raise InvalidDataError("No series name")
        remote_selector = SeriesSelector(self.__source_name, selector.name)
        metadata = self.__client.get_metadata(remote_selector)
        metadata.series = selector
        return metadata

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Get data from the Flight service."""
        remote_selector = SeriesSelector(self.__source_name, selector.name)
        return self.__client.get_data(remote_selector, start_date, end_date)



"""Generic exceptions for Kukur data sources."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0


class InvalidDataError(Exception):
    """Raised when the data itself is invalid."""

    def __init__(self, message: str):
        Exception.__init__(self, f"invalid data: {message}")


class InvalidSourceException(Exception):
    """Raised when the source configuration is invalid."""

    def __init__(self, message: str):
        Exception.__init__(self, f"invalid source: {message}")


class UnknownSourceException(Exception):
    """Raised when the source is not known."""

    def __init__(self, source_name: str):
        Exception.__init__(self, f"source does not exist: {source_name}")



"""
feather contains the Feather data source for Timeseer.

Three formats are supported:
- row based, with many series in one file containing one row per data point
- directory based, with one file per series
- pivot, with many series as columns in one file
"""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0

from .feather import from_config

__all__ = ["from_config"]



"""Common logic for the file formats supported by pyarrow."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Generator

import pyarrow as pa
import pyarrow.compute
import pyarrow.types

from kukur import Metadata, SeriesSelector
from kukur.exceptions import InvalidDataError
from kukur.loader import Loader
from kukur.source.quality import QualityMapper


class BaseArrowSource(ABC):
    """Base class for pyarrow file format data sources."""

    __loader: Loader
    __data_format: str
    __quality_mapper: QualityMapper

    def __init__(self, data_format: str, loader: Loader, quality_mapper: QualityMapper):
        """Create a new data source."""
        self.__loader = loader
        self.__data_format = data_format
        self.__quality_mapper = quality_mapper

    @abstractmethod
    def read_file(self, file_like) -> pa.Table:
        """Read the given file-like object using the pyarrow format reader."""
        ...

    @abstractmethod
    def get_file_extension(self) -> str:
        """Return the file extension for the supported pyarrow format."""
        ...

    def search(self, selector: SeriesSelector) -> Generator[Metadata, None, None]:
        """ArrowSource does not support searching for time series."""

    # pylint: disable=no-self-use
    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Feather currently always returns empty metadata."""
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Read data in one of the predefined formats.

        The complete file will be loaded in an Arrow table during processing.
        """
        data = self.__read_all_data(selector)
        # pylint: disable=no-member
        on_or_after = pyarrow.compute.greater_equal(data["ts"], pa.scalar(start_date))
        before = pyarrow.compute.less(data["ts"], pa.scalar(end_date))
        return data.filter(pyarrow.compute.and_(on_or_after, before))

    def __read_all_data(self, selector: SeriesSelector) -> pa.Table:
        if self.__data_format == "pivot":
            return self._read_pivot_data(selector)

        if self.__data_format == "dir":
            return self._read_directory_data(selector)

        return self._read_row_data(selector)

    def _read_pivot_data(self, selector: SeriesSelector) -> pa.Table:
        all_data = self.read_file(self.__loader.open())
        if selector.name not in all_data.column_names:
            raise InvalidDataError(f'column "{selector.name}" not found')
        data = all_data.select([0, selector.name]).rename_columns(["ts", "value"])
        schema = pa.schema(
            [
                ("ts", pa.timestamp("us", "utc")),
                ("value", _get_value_schema_type(data)),
            ]
        )
        return data.cast(schema)

    def _read_row_data(self, selector: SeriesSelector) -> pa.Table:
        columns = ["series name", "ts", "value"]
        if self.__quality_mapper.is_present():
            columns.append("quality")

        all_data = self.read_file(self.__loader.open()).rename_columns(columns)
        schema = pa.schema(
            [
                ("ts", pa.timestamp("us", "utc")),
                ("value", _get_value_schema_type(all_data)),
            ]
        )
        if self.__quality_mapper.is_present():
            schema = schema.append(pa.field("quality", pa.int8()))
            kukur_quality_values = self._map_quality(all_data["quality"])
            all_data = all_data.set_column(
                3, "quality", pa.array(kukur_quality_values, type=pa.int8())
            )

        # pylint: disable=no-member
        data = all_data.filter(
            pyarrow.compute.equal(all_data["series name"], pa.scalar(selector.name))
        ).drop(["series name"])

        return data.cast(schema)

    def _read_directory_data(self, selector: SeriesSelector) -> pa.Table:
        columns = ["ts", "value"]
        if self.__quality_mapper.is_present():
            columns.append("quality")

        data = self.read_file(
            self.__loader.open_child(f"{selector.name}.{self.get_file_extension()}")
        ).rename_columns(columns)
        schema = pa.schema(
            [
                ("ts", pa.timestamp("us", "utc")),
                ("value", _get_value_schema_type(data)),
            ]
        )
        if self.__quality_mapper.is_present():
            schema = schema.append(pa.field("quality", pa.int8()))
            kukur_quality_values = self._map_quality(data["quality"])
            data = data.set_column(
                2, "quality", pa.array(kukur_quality_values, type=pa.int8())
            )
        return data.cast(schema)

    def _map_quality(self, quality_data: pa.array) -> pa.Table:
        return [
            self.__quality_mapper.from_source(source_quality_value.as_py())
            for source_quality_value in quality_data
        ]


def _get_value_schema_type(data: pa.Table):
    value_type = pa.float64()
    if len(data) > 0:
        if pyarrow.types.is_string(data["value"][0].type):
            value_type = pa.string()
    return value_type



"""Base classes for connections to SQL-like sources from Kukur."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import logging

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timezone, tzinfo
from typing import Dict, Generator, List, Optional, Union

import dateutil.parser
import pyarrow as pa

from kukur import Dictionary, Metadata, SeriesSelector
from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper

logger = logging.getLogger(__name__)


class InvalidMetadataError(Exception):
    """Raised when the metadata is invalid."""

    def __init__(self, message: str):
        Exception.__init__(self, f"invalid metadata: {message}")


@dataclass
class SQLConfig:  # pylint: disable=too-many-instance-attributes
    """Configuration settings for a SQL connection."""

    connection_string: str
    query_string_parameters: bool = False
    list_query: Optional[str] = None
    list_columns: List[str] = field(default_factory=list)
    metadata_query: Optional[str] = None
    metadata_columns: List[str] = field(default_factory=list)
    dictionary_query: Optional[str] = None
    data_query: Optional[str] = None
    data_query_datetime_format: Optional[str] = None
    data_timezone: Optional[tzinfo] = None
    data_query_timezone: Optional[tzinfo] = None
    enable_trace_logging: bool = False

    @classmethod
    def from_dict(cls, data):
        """Create a new SQL data source from a configuration dict."""
        if "connection_string" in data:
            connection_string = data["connection_string"]
        else:
            with open(data["connection_string_path"]) as f:
                connection_string = f.read().strip()

        config = SQLConfig(connection_string)

        config.list_query = data.get("list_query")
        if config.list_query is None and "list_query_path" in data:
            with open(data["list_query_path"]) as f:
                config.list_query = f.read()
        config.list_columns = data.get("list_columns", [])
        config.metadata_query = data.get("metadata_query")
        if config.metadata_query is None and "metadata_query_path" in data:
            with open(data["metadata_query_path"]) as f:
                config.metadata_query = f.read()
        config.metadata_columns = data.get("metadata_columns", [])
        config.dictionary_query = data.get("dictionary_query")
        if config.dictionary_query is None and "dictionary_query_path" in data:
            with open(data["dictionary_query_path"]) as f:
                config.dictionary_query = f.read()
        config.data_query = data.get("data_query")
        if config.data_query is None and "data_query_path" in data:
            with open(data["data_query_path"]) as f:
                config.data_query = f.read()

        config.data_query_datetime_format = data.get("data_query_datetime_format")
        config.query_string_parameters = data.get("query_string_parameters", False)
        if "data_timezone" in data:
            config.data_timezone = dateutil.tz.gettz(data.get("data_timezone"))
        if "data_query_timezone" in data:
            config.data_query_timezone = dateutil.tz.gettz(
                data.get("data_query_timezone")
            )
        if "enable_trace_logging" in data:
            config.enable_trace_logging = data.get("enable_trace_logging", False)

        return config


class BaseSQLSource(ABC):
    """A SQL data source.

    Subclasses should implement a connect() method.
    """

    _config: SQLConfig
    _metadata_value_mapper: MetadataValueMapper
    _quality_mapper: QualityMapper

    def __init__(
        self,
        config: SQLConfig,
        metadata_value_mapper: MetadataValueMapper,
        quality_mapper: QualityMapper,
    ):
        self._config = config
        self._metadata_value_mapper = metadata_value_mapper
        self._quality_mapper = quality_mapper

    def search(
        self, selector: SeriesSelector
    ) -> Generator[Union[SeriesSelector, Metadata], None, None]:
        """Search for time series matching the given selector."""
        if self._config.list_query is None:
            return
        if len(self._config.list_columns) == 0:
            for result in self.__search_names(selector):
                yield result
            return
        for metadata in self.__search_metadata(selector):
            yield metadata

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Read metadata from the DB-API connection."""
        metadata = Metadata(selector)
        if self._config.metadata_query is None:
            return metadata
        connection = self.connect()
        cursor = connection.cursor()

        query = self._config.metadata_query
        params = [selector.name]
        if self._config.query_string_parameters:
            query = query.format(*params)
            params = []
        cursor.execute(query, params)
        row = cursor.fetchone()
        if row:
            for i, name in enumerate(self._config.metadata_columns):
                value = row[i]
                if value is None:
                    continue
                if isinstance(value, str) and value == "":
                    continue
                metadata.set_field(
                    name, self._metadata_value_mapper.from_source(name, value)
                )
        if metadata.dictionary_name is not None:
            metadata.dictionary = self.__query_dictionary(
                cursor, metadata.dictionary_name
            )
        return metadata

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data using the specified DB-API query."""
        if self._config.data_query is None:
            return pa.Table.from_pydict({"ts": [], "value": [], "quality": []})
        connection = self.connect()
        cursor = connection.cursor()

        query = self._config.data_query
        params = [
            selector.name,
            self.__format_date(start_date),
            self.__format_date(end_date),
        ]
        if self._config.query_string_parameters:
            query = query.format(*params)
            params = []

        cursor.execute(query, params)

        timestamps = []
        values = []
        qualities = []
        for row in cursor:
            if self._config.enable_trace_logging:
                logger.info(
                    'Data from "%s (%s)" at %s has value %s with type %s',
                    selector.source,
                    selector.name,
                    row[0],
                    row[1],
                    type(row[1]),
                )
            if isinstance(row[0], datetime):
                ts = row[0]
            else:
                ts = dateutil.parser.parse(row[0])
            if self._config.data_timezone:
                ts = ts.replace(tzinfo=self._config.data_timezone)
            ts = ts.astimezone(timezone.utc)
            value = row[1]
            if value is None:
                value = float("nan")
            if isinstance(value, bytes):
                continue
            if isinstance(value, datetime):
                value = value.isoformat()
            if self._quality_mapper.is_present():
                quality = self._quality_mapper.from_source(row[2])
                qualities.append(quality)
            timestamps.append(ts)
            values.append(value)
        if self._quality_mapper.is_present():
            return pa.Table.from_pydict(
                {"ts": timestamps, "value": values, "quality": qualities}
            )
        return pa.Table.from_pydict({"ts": timestamps, "value": values})

    def __search_names(
        self, selector: SeriesSelector
    ) -> Generator[SeriesSelector, None, None]:
        connection = self.connect()
        cursor = connection.cursor()
        cursor.execute(self._config.list_query)

        for (series_name,) in cursor:
            yield SeriesSelector(selector.source, series_name)

    def __search_metadata(
        self, selector: SeriesSelector
    ) -> Generator[Metadata, None, None]:
        connection = self.connect()
        cursor = connection.cursor()
        dictionary_cursor = None
        if self._config.dictionary_query is not None:
            dictionary_cursor = self.connect().cursor()

        cursor.execute(self._config.list_query)
        series_name_index = None
        for i, name in enumerate(self._config.list_columns):
            if name == "series name":
                series_name_index = i
        if series_name_index is None:
            raise InvalidMetadataError('column "series name" not found')
        for row in cursor:
            selector = SeriesSelector(selector.source, row[series_name_index])
            metadata = Metadata(selector)
            for i, name in enumerate(self._config.list_columns):
                if i == series_name_index:
                    continue
                value = row[i]
                if value is None:
                    continue
                if isinstance(value, str) and value == "":
                    continue
                metadata.set_field(
                    name, self._metadata_value_mapper.from_source(name, value)
                )
            if metadata.dictionary_name is not None and dictionary_cursor is not None:
                metadata.dictionary = self.__query_dictionary(
                    dictionary_cursor, metadata.dictionary_name
                )
            yield metadata

    def __query_dictionary(self, cursor, dictionary_name: str) -> Optional[Dictionary]:
        if self._config.dictionary_query is None:
            return None
        mapping: Dict[int, str] = {}

        query = self._config.dictionary_query
        params = [dictionary_name]
        if self._config.query_string_parameters:
            query = query.format(*params)
            params = []
        cursor.execute(query, params)
        for row in cursor.fetchall():
            mapping[int(row[0])] = row[1]
        if len(mapping) == 0:
            return None
        return Dictionary(mapping)

    def __format_date(self, date):
        if self._config.data_query_timezone:
            date = date.astimezone(self._config.data_query_timezone).replace(
                tzinfo=None
            )
        if self._config.data_query_datetime_format is not None:
            return date.strftime(self._config.data_query_datetime_format)
        return date

    @abstractmethod
    def connect(self):
        """Create a new PEP-249 connection."""
        ...



"""Persistence for Kukur"""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from typing import Optional

from .api_key import ApiKeyRepository

from .base import MigrationRunner  # noqa: F401


class RepositoryRegistry:  # pylint: disable=too-few-public-methods
    """RepositoryRegistry provides access to the different repositories in Kukur."""

    __data_dir: Optional[Path]
    __connection_string: Optional[str] = None

    def __init__(self, *, data_dir: Path = None, connection_string: str = None):
        self.__data_dir = data_dir
        self.__connection_string = connection_string

    def api_key(self) -> ApiKeyRepository:
        """Return the repository used to store api keys."""
        return ApiKeyRepository(self._get_connection_string("api_key"))

    def _get_connection_string(self, name: str) -> str:
        if self.__data_dir is None:
            return f"file:{name}.sqlite"
        return "file:" + str(self.__data_dir / f"{name}.sqlite")



"""Connections to InfluxDB data sources from Timeseer."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Any, Dict, List, Generator, Tuple

import dateutil.parser
import pyarrow as pa

try:
    from influxdb import InfluxDBClient
    from influxdb.exceptions import InfluxDBClientError

    HAS_INFLUX = True
except ImportError:
    HAS_INFLUX = False

from kukur import Metadata, SeriesSelector
from kukur.exceptions import InvalidDataError


class InfluxdbNotInstalledError(Exception):
    """Raised when the influxdb is not available."""

    def __init__(self):
        Exception.__init__(self, "the influxdb is not available. Install influxdb.")


class InvalidClientConnection(Exception):
    """Raised when an error occured when making the connection."""

    def __init__(self, message: str):
        Exception.__init__(self, f"Connection error: {message}")


def from_config(config: Dict[str, Any]):
    """Create a new Influx data source"""
    if not HAS_INFLUX:
        raise InfluxdbNotInstalledError()
    host = config.get("host", "localhost")
    port = config.get("port", 8086)
    ssl = config.get("ssl", False)
    database = config["database"]
    username = config.get("username", "")
    password = config.get("password", "")
    return InfluxSource(host, port, ssl, database, username=username, password=password)


class InfluxSource:
    """An InfluxDB data source."""

    if HAS_INFLUX:
        __client: InfluxDBClient

    def __init__(
        self,
        host: str,
        port: int,
        ssl: bool,
        database: str,
        *,
        username: str,
        password: str,
    ):
        if not HAS_INFLUX:
            raise InfluxdbNotInstalledError()
        try:
            if username != "" and password != "":
                self.__client = InfluxDBClient(
                    host=host, port=port, ssl=ssl, username=username, password=password
                )
            else:
                self.__client = InfluxDBClient(host=host, port=port, ssl=ssl)

            self.__client.switch_database(database)
        except InfluxDBClientError as err:
            raise InvalidClientConnection(err) from err

    def search(self, selector: SeriesSelector) -> Generator[Metadata, None, None]:
        """Search for series matching the given selector."""
        many_series = self.__client.get_list_series()
        fields = self.__client.query("SHOW FIELD KEYS")
        for series in many_series:
            series_name = series.replace("\\", "")
            measurement = series_name.split(",")[0]
            for field in fields.get_points(measurement=measurement):
                yield Metadata(
                    SeriesSelector(
                        selector.source, f'{series_name}::{field["fieldKey"]}'
                    )
                )

    # pylint: disable=no-self-use
    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Influx currently always returns empty metadata."""
        return Metadata(selector)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Return data for the given time series in the given time period."""
        if selector.name is None:
            raise InvalidDataError("No series name")
        measurement, tags, field_key = _parse_influx_series(selector.name)

        query = f"""SELECT time, "{_escape(field_key)}"
                    FROM "{_escape(measurement)}"
                    WHERE time >= $start_date and time <= $end_date"""

        bind_params = {
            "start_date": start_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "end_date": end_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
        }

        for i, (tag_key, tag_value) in enumerate(tags):
            bind_params[str(i)] = tag_value
            query = query + f' and "{_escape(tag_key)}" = ${str(i)}'

        timestamps = []
        values = []
        for item in self.__client.query(
            query=query, bind_params=bind_params
        ).get_points():
            timestamps.append(dateutil.parser.parse(item["time"]))
            values.append(item[field_key])

        return pa.Table.from_pydict({"ts": timestamps, "value": values})


def _parse_influx_series(series: str) -> Tuple[str, List[List[str]], str]:
    field_split = series.rsplit("::", 1)
    field_key = field_split[1]

    measurement_split = field_split[0].split(",")
    measurement = measurement_split[0]
    tags = [tag.split("=") for tag in measurement_split[1:]]
    return measurement, tags, field_key


def _escape(context: str) -> str:
    if '"' in context:
        context = context.replace('"', "")
    if ";" in context:
        context = context.replace(";", "")
    return context



"""Connections to ODBC data sources from Kukur."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

try:
    import pyodbc

    HAS_ODBC = True
except ImportError:
    HAS_ODBC = False

from kukur.source.metadata import MetadataValueMapper
from kukur.source.quality import QualityMapper
from kukur.source.sql import BaseSQLSource, SQLConfig


class ODBCNotInstalledError(Exception):
    """Raised when the pyodbc module is not available."""

    def __init__(self):
        Exception.__init__(self, "the pyodbc modules is not available. Install pyodbc")


def from_config(
    data, metadata_value_mapper: MetadataValueMapper, quality_mapper: QualityMapper
):
    """Create a new ODBC data source from a configuration dict."""
    if not HAS_ODBC:
        raise ODBCNotInstalledError()

    config = SQLConfig.from_dict(data)

    return ODBCSource(config, metadata_value_mapper, quality_mapper)


class ODBCSource(BaseSQLSource):
    """An ODBC data source."""

    def __init__(
        self,
        config: SQLConfig,
        metadata_value_mapper: MetadataValueMapper,
        quality_mapper: QualityMapper,
    ):
        super().__init__(config, metadata_value_mapper, quality_mapper)
        if not HAS_ODBC:
            raise ODBCNotInstalledError()

    def connect(self):
        return pyodbc.connect(self._config.connection_string)



"""Connections to InfluxDB data sources from Timeseer."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from .influxdb import from_config

__all__ = ["from_config"]



"""
csv contains the CSV data source for Timeseer.

Three formats are supported:
- row based, with many series in one file containing one line per data point
- directory based, with one file per series
- pivot, with many series as columns in one file
"""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import csv

from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Generator, Optional

import pyarrow as pa
import pyarrow.csv
import pyarrow.compute

from kukur import Dictionary, Metadata, SeriesSelector

from kukur.loader import Loader, from_config as loader_from_config
from kukur.exceptions import InvalidDataError, InvalidSourceException
from kukur.source.metadata import MetadataMapper, MetadataValueMapper
from kukur.source.quality import QualityMapper


class InvalidMetadataError(Exception):
    """Raised when the metadata is invalid."""

    def __init__(self, message: str):
        Exception.__init__(self, f"invalid metadata: {message}")


def from_config(
    config: Dict[str, str],
    metadata_mapper: MetadataMapper,
    metadata_value_mapper: MetadataValueMapper,
    quality_mapper: QualityMapper,
):
    """Create a new CSV data source from the given configuration dictionary."""
    loaders = CSVLoaders()
    if "path" in config:
        loaders.data = loader_from_config(config, "path", files_as_path=True)
    if "metadata" in config:
        loaders.metadata = loader_from_config(config, "metadata", "r")
    if "dictionary_dir" in config:
        loaders.dictionary = loader_from_config(config, "dictionary_dir", "r")
    data_format = config.get("format", "row")
    return CSVSource(
        data_format, loaders, metadata_mapper, metadata_value_mapper, quality_mapper
    )


@dataclass
class CSVLoaders:
    """Data loaders for CSV sources."""

    data: Optional[Loader] = None
    metadata: Optional[Loader] = None
    dictionary: Optional[Loader] = None


class CSVSource:
    """A CSV data source."""

    __loaders: CSVLoaders
    __data_format: str
    __metadata_mapper: MetadataMapper
    __metadata_value_mapper: MetadataValueMapper
    __quality_mapper: QualityMapper

    def __init__(
        self,
        data_format: str,
        loaders: CSVLoaders,
        metadata_mapper: MetadataMapper,
        metadata_value_mapper: MetadataValueMapper,
        quality_mapper: QualityMapper,
    ):  # pylint:disable=too-many-arguments
        """Create a new CSV data source."""
        self.__loaders = loaders
        self.__data_format = data_format
        self.__metadata_mapper = metadata_mapper
        self.__metadata_value_mapper = metadata_value_mapper
        self.__quality_mapper = quality_mapper

    def search(self, selector: SeriesSelector) -> Generator[Metadata, None, None]:
        """Search for series matching the given selector."""
        if self.__loaders.metadata is None:
            return

        with self.__loaders.metadata.open() as metadata_file:
            reader = csv.DictReader(metadata_file)
            for row in reader:
                if self.__metadata_mapper.from_kukur("series name") not in row:
                    raise InvalidMetadataError('column "series name" not found')
                series_name = row[self.__metadata_mapper.from_kukur("series name")]
                metadata = None
                if selector.name is not None:
                    if series_name == selector.name:
                        metadata = Metadata(
                            SeriesSelector(selector.source, series_name)
                        )
                else:
                    metadata = Metadata(SeriesSelector(selector.source, series_name))

                if metadata is not None:
                    for field, _ in metadata:
                        if self.__metadata_mapper.from_kukur(field) in row:
                            try:
                                value = row[self.__metadata_mapper.from_kukur(field)]
                                metadata.set_field(
                                    field,
                                    self.__metadata_value_mapper.from_source(
                                        field, value
                                    ),
                                )
                            except ValueError:
                                pass
                    if metadata.dictionary_name is not None:
                        metadata.dictionary = self.__get_dictionary(
                            metadata.dictionary_name
                        )
                    yield metadata

    def get_metadata(self, selector: SeriesSelector) -> Metadata:
        """Read metadata, taking any configured metadata mapping into account."""
        metadata = Metadata(selector)
        if self.__loaders.metadata is None:
            return metadata

        with self.__loaders.metadata.open() as metadata_file:
            reader = csv.DictReader(metadata_file)
            for row in reader:
                if self.__metadata_mapper.from_kukur("series name") not in row:
                    raise InvalidMetadataError('column "series name" not found')
                if (
                    row[self.__metadata_mapper.from_kukur("series name")]
                    != selector.name
                ):
                    continue
                for field, _ in metadata:
                    if self.__metadata_mapper.from_kukur(field) in row:
                        try:
                            value = row[self.__metadata_mapper.from_kukur(field)]
                            metadata.set_field(
                                field,
                                self.__metadata_value_mapper.from_source(field, value),
                            )
                        except ValueError:
                            pass

            if metadata.dictionary_name is not None:
                metadata.dictionary = self.__get_dictionary(metadata.dictionary_name)

        return metadata

    def __get_dictionary(self, set_name: str) -> Optional[Dictionary]:
        if self.__loaders.dictionary is None:
            return None
        if not self.__loaders.dictionary.has_child(f"{set_name}.csv"):
            return None
        with self.__loaders.dictionary.open_child(f"{set_name}.csv") as mapping_file:
            reader = csv.reader(mapping_file)
            mapping = {int(row[0]): row[1] for row in reader}
            return Dictionary(mapping)

    def get_data(
        self, selector: SeriesSelector, start_date: datetime, end_date: datetime
    ) -> pa.Table:
        """Read data in one of the predefined formats.

        The complete CSV file will be loaded in an Arrow table during processing.
        """
        data = self.__read_all_data(selector)
        # pylint: disable=no-member
        on_or_after = pyarrow.compute.greater_equal(data["ts"], pa.scalar(start_date))
        before = pyarrow.compute.less(data["ts"], pa.scalar(end_date))
        return data.filter(pyarrow.compute.and_(on_or_after, before))

    def __read_all_data(self, selector: SeriesSelector) -> pa.Table:
        if self.__loaders.data is None:
            raise InvalidSourceException("No data path configured.")
        if self.__data_format == "pivot":
            return _read_pivot_data(self.__loaders.data, selector)

        if self.__data_format == "dir":
            return self._read_directory_data(self.__loaders.data, selector)

        return self._read_row_data(self.__loaders.data, selector)

    def _read_row_data(self, loader: Loader, selector: SeriesSelector) -> pa.Table:
        columns = ["series name", "ts", "value"]
        if self.__quality_mapper.is_present():
            columns.append("quality")
        read_options = pyarrow.csv.ReadOptions(column_names=columns)
        convert_options = pyarrow.csv.ConvertOptions(
            column_types={"ts": pa.timestamp("us", "utc")},
        )
        all_data = pyarrow.csv.read_csv(
            loader.open(), read_options=read_options, convert_options=convert_options
        )
        if self.__quality_mapper.is_present():
            kukur_quality_values = self._map_quality(all_data["quality"])
            all_data = all_data.set_column(
                3, "quality", pa.array(kukur_quality_values, type=pa.int8())
            )

        # pylint: disable=no-member
        data = all_data.filter(
            pyarrow.compute.equal(all_data["series name"], pa.scalar(selector.name))
        )
        return data.drop(["series name"])

    def _read_directory_data(
        self, loader: Loader, selector: SeriesSelector
    ) -> pa.Table:
        columns = ["ts", "value"]
        if self.__quality_mapper.is_present():
            columns.append("quality")
        read_options = pyarrow.csv.ReadOptions(column_names=columns)
        convert_options = pyarrow.csv.ConvertOptions(
            column_types={"ts": pa.timestamp("us", "utc")},
        )
        all_data = pyarrow.csv.read_csv(
            loader.open_child(f"{selector.name}.csv"),
            read_options=read_options,
            convert_options=convert_options,
        )
        if self.__quality_mapper.is_present():
            kukur_quality_values = self._map_quality(all_data["quality"])
            return all_data.set_column(
                2, "quality", pa.array(kukur_quality_values, type=pa.int8())
            )
        return all_data

    def _map_quality(self, quality_data: pa.array) -> pa.Table:
        return [
            self.__quality_mapper.from_source(source_quality_value.as_py())
            for source_quality_value in quality_data
        ]


def _read_pivot_data(loader: Loader, selector: SeriesSelector) -> pa.Table:
    all_data = pyarrow.csv.read_csv(loader.open())
    if selector.name not in all_data.column_names:
        raise InvalidDataError(f'column "{selector.name}" not found')
    columns = ["ts", "value"]
    schema = pa.schema([("ts", pa.timestamp("us", "utc")), ("value", pa.float64())])
    return all_data.select([0, selector.name]).rename_columns(columns).cast(schema)



"""
feather contains the Feather data source for Timeseer.

Three formats are supported:
- row based, with many series in one file containing one row per data point
- directory based, with one file per series
- pivot, with many series as columns in one file
"""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from typing import Dict

import pyarrow as pa
import pyarrow.feather as feather

from kukur.exceptions import InvalidSourceException
from kukur.loader import from_config as loader_from_config
from kukur.source.arrow import BaseArrowSource
from kukur.source.quality import QualityMapper


def from_config(config: Dict[str, str], quality_mapper: QualityMapper):
    """Create a new Feather data source from the given configuration dictionary."""
    data_format = config.get("format", "row")
    if "path" not in config:
        raise InvalidSourceException('Feather sources require a "path" entry')
    loader = loader_from_config(config, files_as_path=True)
    return FeatherSource(data_format, loader, quality_mapper)


class FeatherSource(BaseArrowSource):
    """A Feather data source."""

    def read_file(self, file_like) -> pa.Table:
        """Read the file_like object as Feather."""
        return feather.read_table(file_like)

    def get_file_extension(self) -> str:
        """Return the default feather file extension."""
        return "feather"



"""CLI for Kukur.

This allows launching Kukur as a service or testing data sources."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
import argparse
import csv
import sys

from dateutil.parser import parse as parse_date

import kukur.logging
import kukur.source.test as test_source

from kukur.app import Kukur
from kukur.config import from_toml
from kukur.flight import (
    JSONFlightServer,
    KukurFlightServer,
    KukurServerAuthHandler,
    KukurServerNoAuthHandler,
)


def parse_args():
    """Parse the command line arguments given to Kukur."""
    parser = argparse.ArgumentParser(description="Start Kukur.")
    parser.add_argument(
        "--config-file", default="Kukur.toml", help="Path to the configuration file"
    )
    subparsers = parser.add_subparsers(dest="action", help="Select the CLI action")

    subparsers.add_parser(
        "flight", help="Enable the Arrow Flight interface (the default)"
    )
    test_parser = subparsers.add_parser("test", help="Test data source connectivity")
    api_key_parser = subparsers.add_parser(
        "api-key", help="Create an api key for the Arrow Flight interface"
    )

    test_subparsers = test_parser.add_subparsers(
        dest="test_action", help="Select the type of data to test"
    )
    search_parser = test_subparsers.add_parser(
        "search", help="List all time series in the source"
    )
    metadata_parser = test_subparsers.add_parser(
        "metadata", help="Display metadata for one time series"
    )
    data_parser = test_subparsers.add_parser(
        "data", help="Display data for one time series"
    )

    search_parser.add_argument(
        "--source",
        required=True,
        metavar="SOURCE NAME",
        help="The name of the data source to test",
    )
    metadata_parser.add_argument(
        "--source",
        required=True,
        metavar="SOURCE NAME",
        help="The name of the data source to test",
    )
    metadata_parser.add_argument(
        "--name",
        required=True,
        metavar="SERIES NAME",
        help="The name of the series to query",
    )
    data_parser.add_argument(
        "--source",
        required=True,
        metavar="SOURCE NAME",
        help="The name of the data source to test",
    )
    data_parser.add_argument(
        "--name",
        required=True,
        metavar="SERIES NAME",
        help="The name of the series to query",
    )
    data_parser.add_argument(
        "--start",
        required=True,
        metavar="START DATE",
        help="The start date of the time period to query",
    )
    data_parser.add_argument(
        "--end",
        required=True,
        metavar="END DATE",
        help="The end date of the time period to query",
    )

    api_key_subparser = api_key_parser.add_subparsers(
        dest="api_key_action", help="Select the action for api keys"
    )
    create_parser = api_key_subparser.add_parser("create", help="Create an api key")
    revoke_parser = api_key_subparser.add_parser("revoke", help="Revoke an api key")
    api_key_subparser.add_parser("list", help="List all api keys")

    create_parser.add_argument(
        "--name", required=True, metavar="API KEY NAME", help="The name of the api key"
    )
    revoke_parser.add_argument(
        "--name", required=True, metavar="API KEY NAME", help="The name of the api key"
    )

    return parser.parse_args()


def _serve(kukur_app: Kukur, server_config):
    service = KukurFlightServer(kukur_app)

    auth_handler = KukurServerAuthHandler(kukur_app)
    if "flight" in server_config and not server_config["flight"].get(
        "authentication", True
    ):
        auth_handler = KukurServerNoAuthHandler()

    server = JSONFlightServer(server_config, auth_handler=auth_handler)
    server.register_action_handler("search", service.search)
    server.register_action_handler("get_metadata", service.get_metadata)
    server.register_get_handler("get_data", service.get_data)
    server.register_action_handler("list_sources", kukur_app.list_sources)
    server.serve()


def _test_source(kukur_app: Kukur, args):
    if args.test_action not in ["search", "metadata", "data"]:
        return

    writer = csv.writer(sys.stdout)

    source_name: str = args.source

    series_name: str
    if args.test_action == "search":
        for row in test_source.search(kukur_app, source_name):
            writer.writerow(row)
    elif args.test_action == "metadata":
        series_name = args.name
        for row in test_source.metadata(kukur_app, source_name, series_name):
            writer.writerow(row)
    elif args.test_action == "data":
        series_name = args.name
        start_date = parse_date(args.start)
        end_date = parse_date(args.end)
        for row in test_source.data(
            kukur_app, source_name, series_name, start_date, end_date
        ):
            writer.writerow(row)


def _api_keys(kukur_app: Kukur, args):
    if args.api_key_action not in ["create", "revoke", "list"]:
        return

    writer = csv.writer(sys.stdout)

    if args.api_key_action == "create":
        api_key = kukur_app.get_api_keys().create(args.name)
        writer.writerow([api_key])
    elif args.api_key_action == "revoke":
        kukur_app.get_api_keys().revoke(args.name)
    elif args.api_key_action == "list":
        api_keys = kukur_app.get_api_keys().list()
        for key in api_keys:
            writer.writerow((key.name, key.creation_date.isoformat()))


def _run():
    args = parse_args()
    config = from_toml(args.config_file)
    kukur.logging.configure(config)
    app = Kukur(config)
    if args.action == "test":
        _test_source(app, args)
    elif args.action == "api-key":
        _api_keys(app, args)
    else:
        _serve(app, config)


if __name__ == "__main__":
    _run()



"""Test connections to Timeseer data sources.

This takes care to not persistently store metadata."""

# SPDX-FileCopyrightText: 2021 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

import logging

from datetime import datetime, timezone
from typing import Any, Generator, List

from dateutil.tz import tzlocal

from kukur import Metadata, SeriesSelector, Source


logger = logging.getLogger(__name__)


def search(source: Source, source_name: str) -> Generator[List[Any], None, None]:
    """Test listing all series (or metadata) in a source."""
    header_printed = False
    logger.info('Searching for time series in "%s"', source_name)
    for result in source.search(SeriesSelector(source_name)):
        if isinstance(result, SeriesSelector):
            if not header_printed:
                yield ["series name"]
                header_printed = True
            yield [result.name]
        else:
            if not header_printed:
                yield _get_metadata_header(result)
                header_printed = True
            yield _get_metadata(result)


def metadata(
    source: Source, source_name: str, series_name: str
) -> Generator[List[Any], None, None]:
    """Test fetching metadata from a source.

    This does not store the metadata."""
    logger.info('Requesting metadata for "%s (%s)"', series_name, source_name)
    result = source.get_metadata(SeriesSelector(source_name, series_name))
    yield _get_metadata_header(result)
    yield _get_metadata(result)


def data(
    source: Source,
    source_name: str,
    series_name: str,
    start_date: datetime,
    end_date: datetime,
) -> Generator[List[Any], None, None]:
    """Test fetching data for a time series."""
    start_date = _make_aware(start_date)
    end_date = _make_aware(end_date)
    logger.info(
        'Requesting data for "%s (%s)" from %s to %s',
        series_name,
        source_name,
        start_date,
        end_date,
    )

    table = source.get_data(
        SeriesSelector(source_name, series_name), start_date, end_date
    )
    if "quality" in table.column_names:
        for ts, value, quality in zip(table["ts"], table["value"], table["quality"]):
            yield [ts.as_py().isoformat(), value.as_py(), quality.as_py()]
    else:
        for ts, value in zip(table["ts"], table["value"]):
            yield [ts.as_py().isoformat(), value.as_py()]


def _get_metadata_header(result: Metadata) -> List[str]:
    return ["series name"] + [k for k, _ in result]


def _get_metadata(result: Metadata) -> List[Any]:
    return [result.series.name] + [v for _, v in result]


def _make_aware(timestamp: datetime) -> datetime:
    if timestamp.tzinfo is None:
        return timestamp.replace(tzinfo=tzlocal()).astimezone(timezone.utc)
    return timestamp
